#import libraries and set parameters for train/test split and cross-validation
import numpy as np
SEED = 1234
np.random.seed(SEED)
RATIO_TEST = 0.25
CV_N = 10
CV_K = 5

import os
import pandas as pd
import matplotlib.pyplot as plt

from sklearn import clone

from sklearn.metrics import mean_squared_error as mse
from sklearn.model_selection import ShuffleSplit, KFold
from sklearn.linear_model import LinearRegression as LR
from matplotlib import cm

# Create function for normalization
def normalize(x):
    x = x-np.mean(x)
    std = np.std(x)
    if std != 0:
        x = x/std
    return(x)

# Create function for bootstrapping
def bootstrap_CI(x, K=0.25, N=100):
    K = int(K*len(x))
    out = []
    for i in range(N):
        idx_select = np.random.permutation(len(x))[:K]
        x_ = x[idx_select]
        out.append(np.mean(x_))
    return(np.percentile(out, 50), np.percentile(out, 5), np.percentile(out, 95))

# Create function for Borda count
def borda_count(X):
    X_rank = np.apply_along_axis(np.argsort, 1, X)
    n_feat = X_rank.shape[1]
    counts = np.zeros(n_feat)
    for i_row in range(X_rank.shape[0]):
        for i_col in range(n_feat):
            counts[X_rank[i_row, i_col]] = counts[X_rank[i_row, i_col]] + n_feat - i_col

    return(np.argsort(counts)[::-1])

#%% select dataset
WAVE = 'wave1'
DATASETNAME = 'Week_UK'

#%% create path
DATASETDIR = os.path.join(ROOT_DATADIR, 'datasets', WAVE, DATASETNAME)
DATAFILE = os.path.join(DATASETDIR, 'data.csv')

#%% LOAD DATA
features = pd.read_csv(DATAFILE, index_col=0)

#%% Remove rows with nans
features.dropna(axis=0, inplace=True)

#%%
features.index = np.arange(features.shape[0])

#%% remove the column of the label
labels = features.pop('label')

#%% select features of interest
features = features[['LoneTot','CASEStot','PHQtot','MildActDiff','MildActTimeDiff',
                     'GADtot','SleepTot','BeliefTot','ModActDiff','CHAOStot',
                     'spq22','RPQtot']]
features_description = ['Perceived Loneliness', 'Empathy', 'Depression', 'Mild Activity Time [days]', 'Mild Activity Time [minutes]',
                        'Anxiety', 'Sleep Quality','Beliefs', 'Moderate Activity Time [days]', 'Living conditions',
                        'Schizotypal traits', 'Agression']

#%% from pandas dataframe to numpy
X = features.values
#normalize
X = np.apply_along_axis(normalize, 0, X)

#make values of label to start from 0
y = labels.values - 3

#%% SPLIT TRAIN / VALIDATION / TEST

#create training test splitter
sss = ShuffleSplit(n_splits=1, test_size=RATIO_TEST, random_state=SEED)

#generate splitting indexes
sss.get_n_splits(X, y)
idx_train, idx_test = next(iter(sss.split(X,y)))

#use splitting indexes to split the dataset
X_train, y_train  = X[idx_train, :], y[idx_train] #these samples will be used to train the model
X_test, y_test = X[idx_test, :], y[idx_test] #these samples will be used to test the model

# create the K-fold splitter
skf = KFold(n_splits=CV_K, shuffle=True)

#%%

#store MSEs
MSEs_CI_train = []
MSEs_CI_valid = []

#% set model template
model_template = LR()

#%
MSEs_internal = []

#store importance
importances_ = []

#EXTERNAL CYCLE (CV_N times)
for i_n in range(CV_N): #repeat CV_N times

#K-FOLD CYCLE
    i_fold = 1
    for idx_tr, idx_vl in skf.split(X_train, y_train): #for each fold

    #create train and validation sets using k-fold splitting indexes
        X_tr, y_tr = X_train[idx_tr, :], y_train[idx_tr]
        X_vl, y_vl = X_train[idx_vl, :], y_train[idx_vl]

    #initialize the model
        model = clone(model_template)

    #train
        model.fit(X_tr, y_tr)

    #extract and save importance
        importance = abs(model.coef_)
        importances_.append(importance)

    #simple prediction
        y_tr_pred = model.predict(X_tr)
        y_vl_pred = model.predict(X_vl)

    #get performance
        mse_tr = mse(y_tr, y_tr_pred)
        mse_vl = mse(y_vl, y_vl_pred)

        print(f'{i_n+1}/{CV_N} - {i_fold}/{CV_K} - MSE train: {mse_tr}; MSE valid: {mse_vl}')

        i_fold+=1
        MSEs_internal.append([mse_tr, mse_vl])

#compute mean MSE_tr and mean MSE_vl
MSEs_internal = np.array(MSEs_internal)
MSEs_CI_train = bootstrap_CI(MSEs_internal[:, 0])
MSEs_CI_valid = bootstrap_CI(MSEs_internal[:, 1])

#%% PLOT RESULTS
MSEs_CI_train = np.array(MSEs_CI_train)
MSEs_CI_valid = np.array(MSEs_CI_valid)

#%% VISUALIZING RANKING OF IMPORTANCE for each cycle
rankings = []
for rank in importances_:
    rankings.append(np.argsort(np.argsort(rank)))
rankings = np.array(rankings)

plt.imshow(rankings.transpose(), cmap = 'RdYlGn')
plt.yticks(np.arange(len(features.columns)), features.columns, ha='right')

#%%
importance_average = np.mean(importances_, axis=0)

#%% Get plot with importances
cmap = cm.get_cmap('RdYlGn')
for i in np.arange(len(importance_average)):
    av_importance = importance_average[i]
    #av_importance_color = (av_importance - np.min(importance_average))/(np.max(importance_average) - np.min(importance_average))
    # print(av_rank_col)
    plt.barh(12-i, av_importance, color = 'darkgray')#cmap(av_importance_color))

plt.yticks(np.arange(len(features.columns), 0, -1), features_description, ha='right')
plt.grid()
plt.xlabel('normalized Gini Importance')

#%% FINAL MODEL TRAINING
borda_best = borda_count(importances_)
print(features.columns[borda_best[::-1]])

#set model
model = LR()

#train
model.fit(X_train, y_train)

#prediction
y_train_pred = model.predict(X_train)
y_test_pred = model.predict(X_test)

#get performance
mse_train = mse(y_train, y_train_pred)
mse_test = mse(y_test, y_test_pred)

print(f'MSE train: {mse_train}; MSE test: {mse_test}')

#%%
plt.subplot(211)
plt.plot(y_train)
plt.plot(y_train_pred)

plt.subplot(212)
plt.plot(y_test)
plt.plot(y_test_pred)

#%% Save performance and importance
np.savetxt(os.path.join(DATASETDIR, 'MSE_CI_train_MLR.txt'), MSEs_CI_train)
np.savetxt(os.path.join(DATASETDIR, 'MSE_CI_valid_MLR.txt'), MSEs_CI_valid)
np.savetxt(os.path.join(DATASETDIR, 'importances_MLR.txt'), np.array(importances_))
